from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.encoding import force_str
from rest_framework import exceptions
from rest_framework.fields import ChoiceField
from rest_framework.request import clone_request

from unicef_restlib.fields import ModelChoiceField, SeparatedReadWriteField


class SeparatedReadWriteFieldMetadata:
    """Mixin for providing correct information about SeparatedReadWriteField.
    """

    def get_field_info(self, field):
        if isinstance(field, SeparatedReadWriteField):
            if field.context['request'].method == 'GET':
                field = field.read_field or field.write_field
            else:
                field = field.write_field

        return super().get_field_info(field)


class FSMTransitionActionMetadataMixin:
    """
    Return list of available FSM transitions.
    """

    def _collect_actions(self, instance):
        actions = []
        attrs = dir(instance)
        for attr in attrs:
            instance_action = getattr(instance, attr, None)
            if instance_action and hasattr(instance_action, '_django_fsm'):
                actions.append(instance_action)
        return actions

    def _get_instance(self, view):
        if hasattr(view, 'kwargs') and view.kwargs and 'pk' in view.kwargs:
            return self.get_object(view.queryset.model, view.kwargs["pk"])

    def get_object(self, model, pk):
        obj = get_object_or_404(model.objects.all(), pk=pk)
        return obj

    def determine_actions(self, request, view):
        actions = super().determine_actions(request, view)

        instance = self._get_instance(view)
        if not instance:
            return actions

        current_state = getattr(instance, 'status', None)
        if current_state is None:
            return actions

        allowed_FSM_transitions = []
        for action in self._collect_actions(instance):
            meta = action._django_fsm

            if meta.has_transition(current_state) and meta.has_transition_perm(
                    instance,
                    current_state,
                    request.user
            ):
                transition = meta.get_transition(current_state)

                name = transition.custom.get('name', transition.name)
                if callable(name):
                    name = name(instance)

                allowed_FSM_transitions.append({
                    'code': action.__name__,
                    'display_name': name
                })

        # Move cancel to the end.
        actions["allowed_FSM_transitions"] = sorted(
            allowed_FSM_transitions, key=lambda a: a['code'] == 'cancel'
        )

        return actions


class CRUActionsMetadataMixin:
    """Return "GET" with readable fields as allowed method.
    """

    def determine_actions(self, request, view):
        """For generic class based views we return information about
        the fields that are accepted for 'PUT' and 'POST' methods.
        """
        actions = {}
        for method in {'PUT', 'POST', 'GET'} & set(view.allowed_methods):
            view.request = clone_request(request, method)

            if hasattr(view, 'action_map'):
                view.action = view.action_map.get(method.lower(), None)

            try:
                # Test global permissions
                if hasattr(view, 'check_permissions'):
                    view.check_permissions(view.request)

                # Test object permissions
                instance = None
                lookup_url_kwarg = view.lookup_url_kwarg or view.lookup_field
                if lookup_url_kwarg in view.kwargs and hasattr(view, 'get_object'):
                    instance = view.get_object()

            except (exceptions.APIException, PermissionDenied, Http404):
                pass

            else:
                # If user has appropriate permissions for the view, include
                # appropriate metadata about the fields that should be
                # supplied.
                try:
                    serializer = view.get_serializer(instance=instance)
                except exceptions.PermissionDenied:
                    pass

                else:
                    actions[method] = self.get_serializer_info(serializer)

                    if not actions[method]:
                        del actions[method]

            finally:
                view.request = request

        return actions


class ReadOnlyFieldWithChoicesMixin:
    """Return choices for read only fields.
    """

    def get_field_info(self, field):
        field_info = super().get_field_info(field)
        if isinstance(field, ChoiceField) and hasattr(field, 'choices'):
            field_info['choices'] = [
                {
                    'value': choice_value,
                    'display_name': force_str(choice_name, strings_only=True)
                }
                for choice_value, choice_name in field.choices.items()
            ]
        return field_info


class ModelChoiceFieldMixin:
    """
    Mixin for displaying field choices based on model data.
    """

    def get_field_info(self, field):
        field_info = super().get_field_info(field)
        if (not field_info.get('read_only') and isinstance(field, ModelChoiceField) and hasattr(field, 'choices')):
            field_info['choices'] = [
                {
                    'value': choice_value,
                    'display_name': force_str(choice_name, strings_only=True)
                }
                for choice_value, choice_name in field.choices.items()
            ]
        return field_info



from collections import OrderedDict

from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from django.db.models.fields import related, related_descriptors
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from rest_framework.fields import get_attribute
from rest_framework.serializers import SerializerMetaclass
from rest_framework.utils import model_meta
from rest_framework.validators import BaseUniqueForValidator, UniqueTogetherValidator, UniqueValidator
from unicef_djangolib.fields import CodedGenericRelation

from unicef_restlib.utils import pop_keys


class PKSerializerMixin:
    _pk_field = None

    @property
    def pk_field(self):
        if self._pk_field:
            return self._pk_field

        assert hasattr(self, 'Meta'), (
            'Class {serializer_class} missing "Meta" attribute'.format(
                serializer_class=self.__class__.__name__
            )
        )
        assert hasattr(self.Meta, 'model'), (
            'Class {serializer_class} missing "Meta.model" attribute'.format(
                serializer_class=self.__class__.__name__
            )
        )
        if model_meta.is_abstract_model(self.Meta.model):
            raise ValueError(
                'Cannot use ModelSerializer with Abstract Models.'
            )

        model = self.Meta.model
        info = model_meta.get_field_info(model)

        if 'pk' in self.fields:
            self._pk_field = self.fields['pk']
            return self._pk_field

        if info.pk.name in self.fields:
            self._pk_field = self.fields[info.pk.name]
            return self._pk_field

        assert False, "Serializer {serializer_class} doesn't contain primary key field. " \
            'Add `pk` or `{pk_name}` to fields attribute.'.format(
                serializer_class=self.__class__.__name__,
                pk_name=info.pk.name
            )


class DeletableSerializerMixin(metaclass=SerializerMetaclass):
    """Mixin that allow delete object from list through partial update.
    `_delete` field is used to mark object for delete.
    """
    _delete = serializers.BooleanField(
        write_only=True,
        required=False,
        default=False,
    )

    class Meta:
        delete_field = '_delete'

    @property
    def _delete_field(self):
        """
        Shortcut for `_delete` field.
        :return: `_delete` field.
        """
        field = self.fields.get(self.Meta.delete_field, None)

        if field:
            assert len(field.source_attrs) == 1

        return field

    def get_field_names(self, declared_fields, info):
        field_names = super().get_field_names(declared_fields, info)
        if isinstance(self.parent, serializers.ListSerializer):
            # Add `_delete` field to fields list.
            # So we don't need to declare it in every subclass.
            field_names = list(field_names) + [self.Meta.delete_field]
        return field_names

    def create(self, validated_data):
        if self._delete_field and validated_data.pop(self._delete_field.source, False):
            raise serializers.ValidationError({
                self.Meta.delete_field: [
                    _('You can\'t delete not exist object.')
                ]
            })

        return super().create(validated_data)

    def update(self, instance, validated_data):
        if self._delete_field and validated_data.pop(self._delete_field.source, False):
            instance.delete()
            return None

        return super().update(instance, validated_data)


class WritableListSerializer(serializers.ListSerializer):
    """List serializer that allow modify nested objects including
    creation and deleting.
    """
    @method_decorator(transaction.atomic)
    def update(self, instance, validated_data):
        if isinstance(instance, models.Manager):
            instance = instance.all()
        exists_instances = {i.pk: i for i in instance}
        excess_instances_pks = set(exists_instances.keys())

        model = self.child.Meta.model

        result = list()
        errors = list()
        has_error = False
        for data in validated_data:
            errors.append(dict())

            # PK used to detect exists objects.
            try:
                pk = get_attribute(data, self.child.pk_field.source_attrs)
            except KeyError:
                pk = None

            try:

                if pk:
                    if pk not in exists_instances:
                        raise serializers.ValidationError({
                            self.child.pk_field.field_name: _(
                                '{} with pk `{}` doesn\'t exists.'
                            ).format(
                                model._meta.verbose_name.title(),
                                pk
                            ),
                        })

                    if pk not in excess_instances_pks:
                        raise serializers.ValidationError({
                            self.child.pk_field.field_name: _(
                                'Duplication {} with pk `{}`.'
                            ).format(
                                model._meta.verbose_name,
                                pk
                            )
                        })
                    excess_instances_pks.remove(pk)

                    result.append(self.child.update(
                        exists_instances[pk],
                        data
                    ))
                else:
                    result.append(self.child.create(data))

            except serializers.ValidationError as exc:
                has_error = True
                errors[-1] = serializers.as_serializer_error(exc)

        if has_error:
            raise serializers.ValidationError(errors)

        if excess_instances_pks and not getattr(self.root, 'partial', False):
            model._default_manager.filter(pk__in=excess_instances_pks).delete()

        return result


class WritableNestedChildSerializerMixin(PKSerializerMixin):
    """Mixin that allow serializer to create and modify
    related data as nested serializer.
    """
    class Meta:
        list_serializer_class = WritableListSerializer

    def bind(self, field_name, parent):
        super().bind(field_name, parent)

        # For many nested objects we need to determine each object.
        # For is we use pk. So we need to allow client to send it.
        if isinstance(self.parent, self.Meta.list_serializer_class):
            self.pk_field.read_only = False
            self.pk_field.required = False

    def _divide_deferred_validators(self, fields):
        """
        Remove validation that require instance and save for subsequent use.

        Validators like UniqueValidator require instance for correct working.
        So this validators must be called on the stage when instance is
        available.
        """
        for field in fields.values():
            i = 0

            while i < len(field.validators):
                validator = field.validators[i]
                if not isinstance(validator, (
                        UniqueValidator,
                        UniqueTogetherValidator,
                        BaseUniqueForValidator
                )):
                    i += 1
                    continue

                if not hasattr(field, '_deferred_validators'):
                    field._deferred_validators = []

                field._deferred_validators.append(validator)
                field.validators.remove(validator)

    def to_internal_value(self, data):
        self._divide_deferred_validators(self.fields)
        return super().to_internal_value(data)

    def _run_deferred_validators(self, instance, data):
        """
        Use validators saved in `_divide_deferred_validators` method.
        """
        errors = {}
        for field_name, field in self.fields.items():
            if not hasattr(field, '_deferred_validators'):
                continue

            value = data.get(field_name)
            if not value:
                continue

            for validator in field._deferred_validators:
                if hasattr(validator, 'set_context'):
                    validator.set_context(field)

                if hasattr(validator, 'instance') and not validator.instance:
                    validator.instance = instance

                try:
                    validator(value, field)
                except TypeError:
                    raise Exception("Need to upgrade DRF to >= 3.11")
                except serializers.ValidationError as exc:
                    errors[field.field_name] = exc.detail

        if errors:
            raise serializers.ValidationError(errors)

    def create(self, validation_data):
        # We can create child instance through partial update of
        # parent instance.
        # In this case validation allow to skip required fields
        # and we create instance without these fields.
        # To avoid this we make additional validation for required fields.
        if getattr(self.root, 'partial', False):
            fields = self._writable_fields
            errors = dict()
            for field in fields:
                if field.required and field.field_name not in validation_data:
                    try:
                        field.fail('required')
                    except serializers.ValidationError as exc:
                        errors[field.field_name] = exc.detail

            if errors:
                raise serializers.ValidationError(errors)

        self._run_deferred_validators(None, validation_data)

        return super().create(validation_data)

    def update(self, instance, validation_data):
        self._run_deferred_validators(instance, validation_data)
        return super().update(instance, validation_data)


class WritableNestedParentSerializerMixin:
    """Serializer that allow to create and update nested objects.
    """
    @property
    def writable_nested_serializers(self):
        return [
            field_name for field_name, field in self.fields.items()
            if isinstance(field, serializers.BaseSerializer) and not field.read_only
        ]

    def _get_related_model_field(self, nested_serializer):
        """Return model field through that nested serializer relate with
        parent and type of relation.
        :param nested_serializer:
        :return: Tuple contained model field and relation type
        (`forward`, `reverse`).
        """
        assert len(nested_serializer.source_attrs) == 1, "We don't support fields with complex source."

        related_descriptor = get_attribute(
            self.Meta.model,
            nested_serializer.source_attrs
        )

        if isinstance(related_descriptor, related_descriptors.ReverseOneToOneDescriptor):
            return related_descriptor.related.field, 'forward'
        if isinstance(related_descriptor, related_descriptors.ReverseManyToOneDescriptor):
            return related_descriptor.field, 'forward'
        if (isinstance(
                related_descriptor,
                related_descriptors.ForwardManyToOneDescriptor
        ) and isinstance(related_descriptor.field, related.OneToOneField)):
            return related_descriptor.field, 'reverse'

        assert False, "We don't support many to many relation and forward many to one " \
                      "because updating this relation have side effect."

    def _get_related_data(self, instance, nested_serializer):
        """Return value of foreign key and additional fields on the basis of
        which relation between parent and nested serializers is formed.
        :param instance:
        :param nested_serializer:
        :return: Dictionary with fields values.
        """
        related_model_field, relation_type = self._get_related_model_field(
            nested_serializer
        )

        data = {
            lr_field.attname: getattr(instance, fr_field.attname)
            for lr_field, fr_field in related_model_field.related_fields
        }

        if isinstance(related_model_field, GenericRelation):
            data.update({
                related_model_field.content_type_field_name: ContentType.objects.get_for_model(
                    instance
                ),
            })
        if isinstance(related_model_field, CodedGenericRelation):
            data.update({
                related_model_field.code_field: related_model_field.code,
            })

        return data

    def _save_nested_data(self, instance, field, data):
        related_model_field, relation_type = self._get_related_model_field(
            field
        )

        if instance:
            try:
                nested_instance = field.get_attribute(instance)
            except (ObjectDoesNotExist, AttributeError):
                nested_instance = None
        else:
            nested_instance = None

        # Object forced set to null.
        if data is None:
            if nested_instance and relation_type == 'forward':
                nested_instance.delete()
            return None

        if relation_type == 'forward':
            assert instance, "In the case for forward relation between parent and child " \
                             "saving nested serializer require parent instance."
            related_data = self._get_related_data(instance, field)
            if getattr(field, 'many', False):
                data = [OrderedDict(d, **related_data) for d in data]
            else:
                data = OrderedDict(data, **related_data)

        if nested_instance:
            nested_instance = field.update(nested_instance, data)
        else:
            if isinstance(data, related_model_field.related_model):
                nested_instance = data
            else:
                nested_instance = field.create(data)

        return nested_instance

    @method_decorator(transaction.atomic)
    def create(self, validated_data):
        # Separate nested data.
        nested_data, validated_data = pop_keys(
            validated_data,
            self.writable_nested_serializers
        )

        forward_nested_data = {}
        errors = {}
        for field_name, data in nested_data.items():
            field = self.fields[field_name]

            related_model_field, relation_type = self._get_related_model_field(
                field
            )
            if relation_type == 'forward':
                forward_nested_data[field_name] = data
                continue

            try:
                nested_instance = self._save_nested_data(None, field, data)
            except serializers.ValidationError as exc:
                errors[field_name] = exc.detail
            else:
                validated_data[field_name] = nested_instance

        if errors:
            raise serializers.ValidationError(errors)

        instance = super().create(validated_data)

        errors = dict()
        for field_name, data in forward_nested_data.items():
            field = self.fields[field_name]

            try:
                self._save_nested_data(instance, field, data)
            except serializers.ValidationError as exc:
                errors[field_name] = exc.detail

        if errors:
            raise serializers.ValidationError(errors)

        return instance

    @method_decorator(transaction.atomic)
    def update(self, instance, validated_data):
        # Separate nested data.
        nested_data, validated_data = pop_keys(
            validated_data,
            self.writable_nested_serializers
        )

        forward_nested_data = {}
        errors = {}
        for field_name, data in nested_data.items():
            field = self.fields[field_name]

            related_model_field, relation_type = self._get_related_model_field(
                field
            )
            if relation_type == 'forward':
                forward_nested_data[field_name] = data
                continue

            try:
                nested_instance = self._save_nested_data(instance, field, data)
            except serializers.ValidationError as exc:
                errors[field_name] = exc.detail
            else:
                validated_data[field_name] = nested_instance

        if errors:
            raise serializers.ValidationError(errors)

        instance = super().update(instance, validated_data)

        errors = dict()
        for field_name, data in forward_nested_data.items():
            field = self.fields[field_name]

            try:
                self._save_nested_data(instance, field, data)
            except serializers.ValidationError as exc:
                errors[field_name] = exc.detail

        if errors:
            raise serializers.ValidationError(errors)

        return instance


class WritableNestedSerializerMixin(
        DeletableSerializerMixin,
        WritableNestedChildSerializerMixin,
        WritableNestedParentSerializerMixin
):
    class Meta(
            DeletableSerializerMixin.Meta,
            WritableNestedChildSerializerMixin.Meta
    ):
        pass


class UserContextSerializerMixin:
    def get_user(self):
        return self.context.get('user') or self.context.get('request').user


class RecursiveListSerializer(WritableListSerializer):
    def update(self, instance, validated_data):
        if hasattr(self.child, 'proxied'):
            self.child = self.child.proxied
        return super().update(instance, validated_data)



from collections import OrderedDict

from django.utils.translation import gettext_lazy as _
from model_utils import Choices
from rest_framework import serializers
from rest_framework.fields import empty, Field, SkipField
from rest_framework.utils import model_meta
from rest_framework_recursive.fields import RecursiveField

from unicef_restlib.utils import get_attribute_smart


class builtin_field:
    pass


class ModelChoiceField(serializers.PrimaryKeyRelatedField):
    default_error_messages = {
        'does_not_exist': _('Invalid option "{pk_value}" - option is not available.'),
    }

    @property
    def choices(self):
        if hasattr(self._choices, '__call__'):
            self._choices = self._choices()
        return self._choices

    def get_choice(self, obj):
        raise NotImplementedError

    def _choices(self):
        return OrderedDict(map(self.get_choice, self.get_queryset()))


class SeparatedReadWriteField(Field):
    read_field = None
    write_field = None

    def __init__(self, read_field, write_field=builtin_field, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.read_field = read_field
        self.write_field = write_field

        # update fields from kwargs
        for kwarg_name in {'label', } & set(kwargs.keys()):
            setattr(self.read_field, kwarg_name, kwargs[kwarg_name])

            if self.write_field is not builtin_field:
                setattr(self.write_field, kwarg_name, kwargs[kwarg_name])

    def to_representation(self, value):
        return self.read_field.to_representation(value)

    def to_internal_value(self, data):
        return self.write_field.to_internal_value(data)

    def get_validators(self):
        return self.write_field.get_validators()

    def validate_empty_values(self, data):
        """
        Validate empty values, and either:

        * Raise `ValidationError`, indicating invalid data.
        * Raise `SkipField`, indicating that the field should be ignored.
        * Return (True, data), indicating an empty value that should be
          returned without any further validation being applied.
        * Return (False, data), indicating a non-empty value, that should
          have validation applied as normal.
        """
        if data is empty:
            if getattr(self.root, 'partial', False):
                raise SkipField()
            if self.write_field.required:
                self.fail('required')
            return (True, self.get_default())

        if data is None:
            if not self.write_field.allow_null:
                self.fail('null')
            return (True, None)

        return (False, data)

    def _build_field(self):
        model = getattr(self.parent.Meta, 'model')
        depth = getattr(self.parent.Meta, 'depth', 0)
        info = model_meta.get_field_info(model)

        # Determine any extra field arguments and hidden fields that
        # should be included
        extra_kwargs = self.parent.get_extra_kwargs()
        extra_kwargs.update(self._kwargs)
        extra_kwargs, hidden_fields = self.parent.get_uniqueness_extra_kwargs(
            [self.field_name], [self], extra_kwargs
        )
        extra_field_kwargs = {
            key: value for key, value in self._kwargs.items()
            if key not in ['read_field']
        }

        # Determine the serializer field class and keyword arguments.
        field_class, field_kwargs = self.parent.build_field(
            self.field_name, info, model, depth
        )

        # Include any kwargs defined in `Meta.extra_kwargs`
        extra_field_kwargs.update(
            extra_kwargs.get(self.field_name, {})
        )
        field_kwargs = self.parent.include_extra_kwargs(
            field_kwargs, extra_field_kwargs
        )

        # Create the serializer field.
        return field_class(**field_kwargs)

    def bind(self, field_name, parent):
        super().bind(field_name, parent)

        self.read_field.bind(field_name, parent)

        if self.write_field is builtin_field:
            self.write_field = self._build_field()
        self.write_field.bind(field_name, parent)


class WriteListSerializeFriendlyRecursiveField(RecursiveField):
    @property
    def proxied(self):
        self._proxied = super().proxied
        if (self._proxied and not self._proxied.context and self.bind_args[1] and self.bind_args[1].context):
            self._proxied.context = self.bind_args[1].context
        return self._proxied


class CommaSeparatedExportField(serializers.Field):
    export_attr = None

    def __init__(self, *args, **kwargs):
        self.export_attr = kwargs.pop('export_attr', None)
        super().__init__(*args, **kwargs)

    def get_attribute(self, instance):
        try:
            return get_attribute_smart(instance, self.source_attrs)
        except (KeyError, AttributeError) as exc:
            if not self.required and self.default is empty:
                raise SkipField()
            msg = (
                'Got {exc_type} when attempting to get a value for field '
                '`{field}` on serializer `{serializer}`.\nThe serializer '
                'field might be named incorrectly and not match '
                'any attribute or key on the `{instance}` instance.\n'
                'Original exception text was: {exc}.'.format(
                    exc_type=type(exc).__name__,
                    field=self.field_name,
                    serializer=self.parent.__class__.__name__,
                    instance=instance.__class__.__name__,
                    exc=exc
                )
            )
            raise type(exc)(msg)

    def to_representation(self, value):
        value = set(value)

        if self.export_attr:
            value = [get_attribute_smart(item, self.export_attr) for item in value]

        return ', '.join([str(item) for item in value if item])


class DynamicChoicesField(serializers.ChoiceField):
    def __init__(self, *args, **kwargs):
        self._current_choices = {}
        super().__init__(*args, **kwargs)

    @property
    def choices(self):
        return self._current_choices

    @choices.setter
    def choices(self, value):
        self._current_choices = value

    @property
    def choice_strings_to_values(self):
        if isinstance(self.choices, Choices):
            return {k: v for k, v in self.choices}
        return {
            str(key): key for key in self.choices.keys()
        }

    @choice_strings_to_values.setter
    def choice_strings_to_values(self, value):
        # no need to do here anything
        return


class FunctionRelatedField(serializers.RelatedField):

    def __init__(self, callable_function=None, **kwargs):
        assert callable_function is not None, 'The `callable_function` argument is required.'
        self.callable_function = callable_function
        super().__init__(**kwargs)

    def to_representation(self, obj):
        return self.callable_function(obj)



from collections.abc import Iterable, Mapping
from itertools import chain

from django.db.models import Manager, QuerySet
from rest_framework.fields import get_attribute


def pop_keys(d, keys):
    res, rem = dict(), dict()
    for key, value in d.items():
        if key in keys:
            res[key] = value
        else:
            rem[key] = value
    return res, rem


def get_attribute_smart(instance, attrs):
    """A bit smarter version of rest_framework.fields.get_attribute.
    Has ability to work with lists, so it can be used to look deep inside relations.
    Also is suitable for dicts.

    Example usage:

    get_attribute_smart({"instances": [{"id": 1}, {"id": 2}, {"id": 3}]}, "instances.id")
    """

    if instance is None or not attrs:
        return instance

    if isinstance(attrs, str):
        attrs = attrs.split('.')

    if isinstance(instance, (Iterable, QuerySet)) and not isinstance(instance, (Mapping, str)):
        instance = list([get_attribute_smart(obj, [attrs[0]]) for obj in instance])
        if all(map(lambda obj: isinstance(obj, QuerySet), instance)):
            instance = chain(*instance)
    else:
        instance = get_attribute(instance, [attrs[0]])

        if isinstance(instance, Manager):
            instance = instance.all()

    return get_attribute_smart(instance, attrs[1:])



from rest_framework.pagination import PageNumberPagination


class OptionalPaginationMixin:
    def get_page_size(self, request):
        if self.page_size_query_param:
            try:
                if request.query_params[self.page_size_query_param] == 'all':
                    return None
            except KeyError:
                pass
        return super().get_page_size(request)


class DynamicPageNumberPagination(
        OptionalPaginationMixin,
        PageNumberPagination
):
    page_size = 10
    page_size_query_param = 'page_size'
    max_page_size = 1000



from rest_framework_nested import routers


class NestedComplexRouter(routers.NestedSimpleRouter):
    """Set attributes for NestedViewSetMixin.
    Requires correct value for lookup argument,
    instead FieldError will be raised on queryset filtering.
    """

    def register(self, prefix, viewset, basename=None):
        super().register(prefix, viewset, basename=basename)

        # this is run in super()
        # re-running as we need that variables
        parent_registry = [
            registered for registered
            in self.parent_router.registry
            if registered[0] == self.parent_prefix
        ]
        parent_registry = parent_registry[0]
        parent_prefix, parent_viewset, parent_basename = parent_registry

        viewset.parent = parent_viewset
        viewset.parent_lookup_field = self.nest_prefix[:-1]
        viewset.parent_lookup_kwarg = self.nest_prefix + getattr(
            parent_viewset,
            'lookup_field',
            'pk'
        )



#!/usr/bin/env python
import ast
import codecs
import os.path
import re
import sys
from codecs import open

from setuptools import find_packages, setup

ROOT = os.path.realpath(os.path.dirname(__file__))
init = os.path.join(ROOT, 'src', 'unicef_restlib', '__init__.py')
_version_re = re.compile(r'__version__\s+=\s+(.*)')
_name_re = re.compile(r'NAME\s+=\s+(.*)')

sys.path.insert(0, os.path.join(ROOT, 'src'))

with open(init, 'rb') as f:
    content = f.read().decode('utf-8')
    VERSION = str(ast.literal_eval(_version_re.search(content).group(1)))
    NAME = str(ast.literal_eval(_name_re.search(content).group(1)))


def read(*files):
    content = []
    for f in files:
        content.extend(codecs.open(os.path.join(ROOT, 'src', 'requirements', f), 'r').readlines())
    return "\n".join(filter(lambda l:not l.startswith('-'), content))


setup(name=NAME,
      version=VERSION,
      url='https://github.com/unicef/unicef-rest-export',
      author='UNICEF',
      author_email='dev@unicef.org',
      license="Apache 2 License",
      description='Django package that handles exporting of data',
      long_description=codecs.open('README.rst').read(),
      package_dir={'': 'src'},
      packages=find_packages(where='src'),
      include_package_data=True,
      install_requires=(
          'django',
          'django-fsm',
          'djangorestframework',
          'djangorestframework-recursive',
          'drf-nested-routers',
          'unicef-djangolib',
      ),
      extras_require={
          'test': (
              'coverage',
              'django-fsm',
              'factory-boy',
              'faker',
              'flake8',
              'isort',
              'pyflakes',
              'pytest',
              'pytest-cov',
              'pytest-django',
              'pytest-echo',
              'pytest-pythonpath',
          ),
      },
      platforms=['any'],
      classifiers=[
          'Environment :: Web Environment',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Framework :: Django',
          'Framework :: Django :: 2.2',
          'Framework :: Django :: 3.0',
          'Framework :: Django :: 3.1',
          'Framework :: Django :: 3.2',
          'Intended Audience :: Developers'],
      scripts=[],
      )
